import static choco.Choco.*;
import choco.kernel.model.variables.integer.IntegerVariable;
import choco.cp.model.CPModel;
import choco.kernel.model.Model;
import java.util.*;

public class Lecture {

    Course course;
    String name;
    int seqNo;
    IntegerVariable startTime; // {0 .. 79}
    boolean newDay;
    Model model;

    Lecture(Course course,int seqNo){
	this.course = course;
	this.name   = course.name;
	this.seqNo  = seqNo;
	startTime   = makeIntVar(name + "-"+ seqNo,0,79);
	newDay      = false;
	model       = course.model;
	model.addVariable(startTime);
    }
    //
    // NOTE: a time t is in {0 .. 79}
    //       t%8 + 9 gives hour in day
    //      (t/8)%5  gives day
    //       t/40    gives semester (0 or 1)
    //
    //  Semester 0
    //  time: 9 10 11 12 13 14 15 16
    //  Mon:  0  1  2  3  4  5  6  7
    //  Tue:  8  9 10 11 12 13 14 15
    //  Wed: 16 17 18 19 20 21 22 23
    //  Thu: 24 25 26 27 28 29 30 31
    //  Fri: 32 33 34 35 36 37 38 39
    //
    //  Semester 1
    //  time: 9 10 11 12 13 14 15 16
    //  Mon: 40 41 42 43 44 45 46 47
    //  Tue: 48 49 50 51 52 53 54 55
    //  Wed: 56 57 58 59 60 61 62 63
    //  Thu: 64 65 66 67 68 69 70 71
    //  Fri: 72 73 74 75 76 77 78 79
    //

    void sameDay(Lecture lect2){
	IntegerVariable start2 = lect2.startTime;
	model.addConstraint(eq(mod(div(startTime,8),5),mod(div(start2,8),5)));
    }
    //
    // (t/8)%5  gives day
    //

    void differentDay(Lecture lect2){
	IntegerVariable start2 = lect2.startTime;
	model.addConstraint(neq(mod(div(startTime,8),5),mod(div(start2,8),5)));
    }
    //
    // (t/8)%5  gives day
    //

     void sameSemester(Lecture lect2){
	IntegerVariable start2 = lect2.startTime;
	model.addConstraint(eq(div(startTime,40),div(start2,40)));
    }
    //
    // t/40 is semester
    //

     void nextSemester(Lecture lect2){
	IntegerVariable start2 = lect2.startTime;
	model.addConstraint(eq(div(startTime,40),1));
	model.addConstraint(eq(div(start2,40),0));
    }
    //
    // this has to be in semster 1 and lect2 in semester 0
    //

    void meets(Lecture lect2){
	IntegerVariable start2 = lect2.startTime;
	model.addConstraint(eq(1,minus(mod(start2,8),mod(startTime,8))));
	this.sameDay(lect2);
    }
    //
    // t%8 gives hour in day (0 to 8)
    // lecture lect2 starts 1 hour after lect on sameDay
    //

    void before(Lecture lect2){
	IntegerVariable start2 = lect2.startTime;
	model.addConstraint(lt(startTime,start2));
    }
    //
    // this lecture is before lecture lect2
    //

     void notEqual(Lecture lect2){
	IntegerVariable start2 = lect2.startTime;
	model.addConstraint(neq(startTime,start2));
    }
    //
    // this lecture cannot be at same thime as lect
    //

    void setStartTime(int t){
	//System.out.println(startTime.getName() +" = "+ t);
	model.addConstraint(eq(startTime,t));
    }
    //
    // set the start time, when we have pre-set schedules
    //

     void forbiddenTime(int t){
	 //System.out.println(startTime.getName() +" != "+ t);
	model.addConstraint(neq(startTime,t));
    }
    //
    // cannot start at time t
    //
}