import java.util.*;
import java.io.*;
import static choco.Choco.*;
import choco.cp.model.CPModel;
import choco.cp.solver.CPSolver;
import choco.kernel.model.Model;
import choco.kernel.solver.Solver;
import choco.kernel.model.variables.integer.IntegerVariable;
import choco.cp.solver.search.integer.varselector.*;
import choco.cp.solver.search.integer.valselector.*;

public class TimTab {

    ArrayList<Year> years;
    ArrayList<Course> courses;
    Model model;
    Solver solver;

    public TimTab(String fnameCourses, String fnameYears) throws IOException {
	model   = new CPModel();
	courses = new ArrayList<Course>();
	BufferedReader coursesFin = new BufferedReader(new FileReader(fnameCourses));
	String b = coursesFin.readLine();
	boolean moreData = b != null;
	while (moreData){
	    StringTokenizer st = new StringTokenizer(b," ");
	    Course course = new Course(model,st);
	    courses.add(course);
	    moreData = (b = coursesFin.readLine()) != null;	
	}
	coursesFin.close();
	years = new ArrayList<Year>();
	BufferedReader yearsFin = new BufferedReader(new FileReader(fnameYears));
	b = yearsFin.readLine();
	moreData = b != null;
	while (moreData){
	    StringTokenizer st = new StringTokenizer(b," ");
	    String name = st.nextToken();
	    if (!name.equals("//")){
		Year year = new Year(model,name,st,courses);
		years.add(year);
	    }
	    moreData = (b = yearsFin.readLine()) != null;	
	}
	yearsFin.close();
	for (Year year : years) year.init();
    }
    //
    // given a file of courses with lecture patterns (courses.txt)
    // produce an arrayList of courses ... then
    // given a file of years, listing courses in each year (years.txt)
    // produce an arrayList of years
    //

    IntegerVariable[] getDecisionVariables(){
	ArrayList<IntegerVariable> vars = new ArrayList<IntegerVariable>();
	for (Course course : courses)
	    for (Lecture lecture : course.lectures)
		vars.add(lecture.startTime);
	int n = vars.size();
	IntegerVariable[] decVar = new IntegerVariable[n];
	for (int i=0;i<n;i++) decVar[i] = vars.get(i);
	return decVar;
    }


    void imposeFixedTimes(String fname) throws IOException {
	BufferedReader fin = new BufferedReader(new FileReader(fname));
	String b = fin.readLine();
	boolean moreData = b != null;
	while (moreData){
	    StringTokenizer st = new StringTokenizer(b," ");
	    String courseName = st.nextToken();
	    if (!courseName.equals("//")){
		ArrayList<Integer> times = new ArrayList<Integer>();
		Course course = getCourse(courseName);
		while (st.hasMoreTokens()){
		    String sdt   = st.nextToken(); // semseterDayTime
		    int semester = Character.digit(sdt.charAt(0),10);
		    int time1    = Character.digit(sdt.charAt(3),10);
		    int time2    = Character.digit(sdt.charAt(4),10);
		    int day      = -1;
		    if(sdt.substring(1,3).equals("MO")) day = 0;	
		    if(sdt.substring(1,3).equals("TU")) day = 1;	
		    if(sdt.substring(1,3).equals("WE")) day = 2;	
		    if(sdt.substring(1,3).equals("TH")) day = 3;	
		    if(sdt.substring(1,3).equals("FR")) day = 4;	
		    int time = 40*(semester-1) + 8*day + time1*10 + time2 - 9;
		    times.add(time);
		}
		if (times.size() != course.lectures.size()) throw new TimTabException("mismatch times: "+ times +" and lectures: "+ course);
		for (int i=0;i<times.size();i++){
		    Lecture lecture = course.lectures.get(i);
		    int time = times.get(i);
		    lecture.setStartTime(time);
		}
	    }
	    moreData = (b = fin.readLine()) != null;	
	}
	fin.close();
    }

    void forbiddenTimes(String fname) throws IOException {
	BufferedReader fin = new BufferedReader(new FileReader(fname));
	String b = fin.readLine();
	boolean moreData = b != null;
	while (moreData){
	    StringTokenizer st = new StringTokenizer(b," ");
	    String courseName = st.nextToken();
	    Course course = getCourse(courseName);
	    ArrayList<Integer> times = new ArrayList<Integer>();
	    while (st.hasMoreTokens()){
		String sdt   = st.nextToken(); // semseterDayTime
		int semester = Character.digit(sdt.charAt(0),10);
		int time1    = Character.digit(sdt.charAt(3),10);
		int time2    = Character.digit(sdt.charAt(4),10);
		int day      = -1;
		if(sdt.substring(1,3).equals("MO")) day = 0;	
		if(sdt.substring(1,3).equals("TU")) day = 1;	
		if(sdt.substring(1,3).equals("WE")) day = 2;	
		if(sdt.substring(1,3).equals("TH")) day = 3;	
		if(sdt.substring(1,3).equals("FR")) day = 4;	
		int time = 40*(semester-1) + 8*day + time1*10 + time2 - 9;
		times.add(time);
	    }
	    for (int time : times)
		for (Lecture lecture : course.lectures)
		    lecture.forbiddenTime(time);
	    moreData = (b = fin.readLine()) != null;	
	}
	fin.close();
    }

    void lectureClashes(String fname) throws IOException {
	BufferedReader fin = new BufferedReader(new FileReader(fname));
	String b = fin.readLine();
	boolean moreData = b != null;
	while (moreData){
	    StringTokenizer st = new StringTokenizer(b," ");
	    String name = st.nextToken();
	    if (!name.equals("***")){
		Course course = getCourse(name); //st.nextToken());
		ArrayList<Course> clashCourses = new ArrayList<Course>();
		while (st.hasMoreTokens()){
		    String s = st.nextToken();
		    clashCourses.add(getCourse(s));
		}
		for (Course clashCourse : clashCourses)
		    course.noConflicts(clashCourse);
	    }
	    moreData = (b = fin.readLine()) != null;	
	}
	fin.close();
    }
    //
    // courses that cannot occur at same time
    //

    void specialisms (String fname) throws IOException {
	BufferedReader fin = new BufferedReader(new FileReader(fname));
	String b = fin.readLine();
	boolean moreData = b != null;
	while (moreData){
	    StringTokenizer st = new StringTokenizer(b," ");
	    ArrayList<Course> specialism = new ArrayList<Course>();
	    while (st.hasMoreTokens()){
		String name = st.nextToken();
		//System.out.print("("+name +","+ getCourse(name) +") ");
		Course course = getCourse(name);
		if (course == null) throw new TimTabException("no such course: "+ name);
		specialism.add(course);
	    }
	    //System.out.println();
	    for (int i=0;i<specialism.size()-1;i++)
		for (int j=i+1;j<specialism.size();j++){
		    Course c1 = specialism.get(i);
		    if (c1 == null) throw new TimTabException("null specialism c1");
		    Course c2 = specialism.get(j);
		    if (c2 == null) throw new TimTabException("null specialism c2");
		    c1.noConflicts(c2);
		}
	    moreData = (b = fin.readLine()) != null;	
	}
	fin.close();
    }
    //
    // courses that cannot occur at same time
    //

    void prereqs (String fname) throws IOException {
	BufferedReader fin = new BufferedReader(new FileReader(fname));
	String b = fin.readLine();
	boolean moreData = b != null;
	while (moreData){
	    StringTokenizer st = new StringTokenizer(b," ");
	    Course course = getCourse(st.nextToken());
	    while (st.hasMoreTokens()) getCourse(st.nextToken()).prereq(course);
	    moreData = (b = fin.readLine()) != null;	
	}
	fin.close();
    }
    //
    // prerequisites for courses
    // NOTE: use file temp.txt
    //

    Course getCourse(String name){
	for (Course course : courses) 
	    if (course.name.equals(name)) return course;
	return null;
    }

    boolean solve(){
	solver = new CPSolver();
	solver.read(model);
	solver.setVarIntSelector(new MinDomain(solver,solver.getVar(getDecisionVariables())));
	return solver.solve();
    }

    void stats(){
	System.out.println("nodes: "+ solver.getNodeCount() +"   cpu: "+ solver.getTimeCount());
    }

    void printSolution(){
	for (Course course : courses)
	    for (Lecture lect : course.lectures){
		System.out.print(lect.startTime.getName() +" ");
		System.out.println(solver.getVar(lect.startTime).getVal());
	    }
    }

    void printTimetable(ArrayList<Course> courses){
	ArrayList[] bucket = new ArrayList[80];
	for (int i=0;i<80;i++) bucket[i] = new ArrayList<Course>();
	for (Course course : courses)
	    for (Lecture lecture : course.lectures){
		int time = solver.getVar(lecture.startTime).getVal();
		bucket[time].add(lecture);
	    }
	for (int i=0;i<80;i++){
	    int semester = i/40 + 1;
	    int day      = (i/8)%5;
	    int hour     = i%8 + 9;
	    System.out.print(semester);
	    if (day == 0) System.out.print("-Mon-");
	    if (day == 1) System.out.print("-Tue-");
	    if (day == 2) System.out.print("-Wed-");
	    if (day == 3) System.out.print("-Thu-");
	    if (day == 4) System.out.print("-Fri-");
	    System.out.print(hour +" ");
	    for (Object o : bucket[i]){
		Lecture lect = (Lecture) o;
		System.out.print(lect.name +"-"+ lect.seqNo +" ");
	    }
	    System.out.println();
	}	    
    }

    
    void printYear(String name){
	Year year = null;
	for (Year y : years) if (y.name.equals(name)) year = y;
	if (year == null) throw new TimTabException("no such year: "+ name);       
	for (int i=0;i<10;i++) System.out.print(year.name +" -- ");
	System.out.println(year.name);
	printTimetable(year.courses);
    }
    
    
    public static void main(String[] args) throws IOException {

	if (args.length == 0 || args[0].equals("-h")){
	    System.out.println("TimTab version 20150408 \n"+
			       "-courses     FILE courses with lecture pattern for each course (compulsory) \n"+
			       "-years       FILE years with courses that make that year (compulsory) \n"+
			       "-fixed       FILE fixed times for lectures \n" +
			       "-forbidden   FILE times when lectures are forbidden \n" +
                               "-clashes     FILE courses that must take place at different times \n"+
			       "-specialisms FILE courses that make up a specialism (must not clash) \n" +
			       "-prereq      FILE prerequisites (wip) \n" +
                               "-v           verbose \n" +
			       "-h           Print this help message");
	    return;
	}
	
	boolean verbose = false;
	String courses= "", years = "", fixed = "", forbidden = "", clashes = "", specialisms = "", prereq = "";
	for (int i=0;i<args.length;i++){
	    if (args[i].equals("-courses")) courses = args[i+1];
	    if (args[i].equals("-years")) years = args[i+1];
	    if (args[i].equals("-fixed")) fixed = args[i+1];
	    if (args[i].equals("-forbidden")) forbidden = args[i+1];
	    if (args[i].equals("-clashes")) clashes = args[i+1];
	    if (args[i].equals("-specialisms")) specialisms = args[i+1];
	    if (args[i].equals("-prereq")) prereq = args[i+1];
	    if (args[i].equals("-v")) verbose = true;
	}	

	TimTab timTab = new TimTab(courses,years);
	if (!fixed.equals("")) timTab.imposeFixedTimes(fixed);
	if (!forbidden.equals("")) timTab.forbiddenTimes(forbidden);
	if (!clashes.equals("")) timTab.lectureClashes(clashes);
	if (!specialisms.equals("")) timTab.specialisms(specialisms); // are these just clashes (above)?
	if (!prereq.equals("")) timTab.prereqs(prereq); // not clear what these are yet
	System.out.println(timTab.solve());
	timTab.stats();
	//timTab.printSolution();
	//timTab.printYear("L1a");
	if (verbose) timTab.printTimetable(timTab.courses);
    }
}