import java.util.*;
import java.io.*;
import static choco.Choco.*;
import choco.cp.model.CPModel;
import choco.cp.solver.CPSolver;
import choco.kernel.model.Model;
import choco.kernel.solver.Solver;
import choco.kernel.model.variables.integer.IntegerVariable;
import choco.kernel.model.constraints.Constraint;

public class Year {

    String name;
    ArrayList<Course> courses;
    IntegerVariable[][] conflict;
    IntegerVariable[] semester;
    IntegerVariable totalConflicts;
    Model model;

    public Year(Model model, String name, StringTokenizer st, ArrayList<Course> allCourses){
	this.name    = name;
	this.model   = model;
	this.courses = new ArrayList<Course>();
	while (st.hasMoreTokens()){
	    String courseName = st.nextToken();
	    Course course = getCourse(allCourses,courseName);
	    if (course == null) throw new TimTabException("unknown course: "+ courseName);
	    courses.add(course);
	    course.years.add(this);
	}
    }
    //
    // Create a year
    // Given an arrayList of courses, add to the year the courses that  make up that year
    //

    void init(){
	int n = courses.size();
	semester = new IntegerVariable[n];
	for (int i=0;i<n;i++) semester[i] = courses.get(i).semester;
	IntegerVariable count = makeIntVar("count",n/2-1,n/2+1);
	model.addConstraint(eq(count,sum(semester))); // balance across semesters
	conflict = new IntegerVariable[n][n];
	for (int i=0;i<n-1;i++)
	    for (int j=i+1;j<n;j++) conflict[i][j] = conflict[j][i] = makeIntVar("conf-"+ name +"_"+ i +"_"+ j,0,1);
	for (int i=0;i<n;i++) conflict[i][i] = constant(0);
	totalConflicts = makeIntVar("totalConflicts-"+ name,0,0);//n*(n-1));
	for (int i=0;i<n-1;i++)
	    for (int j=i+1;j<n;j++){
		Course course1 = courses.get(i);
		Course course2 = courses.get(j);
		Constraint c = noConflicts(course1,course2);
		model.addConstraint(ifOnlyIf(eq(conflict[i][j],0),c));
	    }
	IntegerVariable[] flatConflicts = new IntegerVariable [n*n];
	for (int i=0,k=0;i<n;i++)
	    for (int j=0;j<n;j++,k++) flatConflicts[k] = conflict[i][j];
	model.addConstraint(eq(totalConflicts,sum(flatConflicts)));
    }
    //
    // after adding all courses to a year we can initialise it
    // creating the conflict array and posting a constraint on 
    // the occurrences of semester 0 and semester 1 courses
    // (global cardinality constraint use for balance)
    //

    Course getCourse(ArrayList<Course> allCourses, String name){
	for (Course course : allCourses) 
	    if (course.name.equals(name)) return course;
	return null;
    }
    //
    // given a name (string) and an arrayList of courses, delive that course
    //

    Constraint noConflicts(Course course1, Course course2){
	Constraint c = TRUE;
	for (Lecture lect1 : course1.lectures)
	    for (Lecture lect2 : course2.lectures)
		c = and(c,neq(lect1.startTime,lect2.startTime));
	return c;
    }
}